#!/usr/bin/env python2.7
import sys,os
import sqlite3
import argparse
from .MMlib3 import bash, printerr
from .ncbi_lib import load_ncbi_config, save_ncbi_config
from pathlib import Path

def main():

  userfile=str(Path.home())+'/.ncbi_taxdb'

  parser = argparse.ArgumentParser(description='Generate a tax.db for ncbi_taxonomy_tree.\nWith no arguments, the program downloads the current "taxdump" from ncbi taxonomy and generates the tax.db in the current directory. It also writes a ~/.ncbi_taxdb file so that later ncbi_taxonomy_tree will automatically find tax.db', prog='ncbi_taxonomy_tree -makedb')
  parser.add_argument("-f","--folder",  help="folder containing names.dmp and nodes.dmp files from ncbi. tax.db will be created there")
  parser.add_argument("-o","--output",  help="folder where tax.db will be created. Also, if --folder is not provided, files will be downloaded here")  
  parser.add_argument("-s","--skip",  help="skip saving the DB location as default (in file ~/.ncbi_config)")
  #parser.add_argument("-w","--overwrite",  help="overwrite without prompting the .ncbi_taxdb file in your home folder, if existing")

  args = parser.parse_args()

  download_files=False
  if not args.folder:
    args.folder='./'
    download_files=True
  if not args.output:
    args.output=args.folder

  print(f'Working directory: {args.output}')
  os.chdir(args.output)
    
  if download_files:
      print( 'Downloading taxdump.tar.gz...', end='', flush=True)
      b = bash('wget ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz')
      if b[0]: raise Exception(b[1])
      b = bash('wget ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz.md5')
      if b[0]: raise Exception(b[1])
      md5 = [ x.rstrip() for x in open('taxdump.tar.gz.md5') ][0]
      b = bash('md5sum taxdump.tar.gz')
      if b[0]: raise Exception(b[1])
      if not md5 == b[1]:
          raise Exception("ERROR: something went wrong with wget, md5sums do not match: %s != %s" % (md5,b[1]))
      b=bash('gunzip -c taxdump.tar.gz | tar xf -')
      if b[0]: raise Exception(b[1])
      args.folder = '.'
      print( 'DONE!')
      


  args.folder = args.folder.rstrip('/') +'/'

  names_dict = {}
  for i in open(args.folder + 'names.dmp'):
      tax_id, name_txt, unique_name, name_class = i.rstrip('\t|\n').split('\t|\t')
      names_dict.setdefault(tax_id, {}).setdefault(name_class, name_txt)

  nodes_dict = {}
  for i in open(args.folder +'nodes.dmp'):
      tax_id, parent_tax_id, rank= i.split('\t|\t')[:3]
      nodes_dict.setdefault(tax_id, {}).setdefault('parent_tax_id', parent_tax_id)
      nodes_dict.setdefault(tax_id, {}).setdefault('rank', rank)

  print(f"Preparing to write file: {args.output.rstrip('/')+'/tax.db'}")

  print("I'm going to insert %s taxids into tax.db, this may take a while\n" % len(nodes_dict))
  con = sqlite3.connect(args.output + 'tax.db')
  with con:
      cur = con.cursor()    
      cur.execute("DROP TABLE IF EXISTS species")
      cur.execute("CREATE TABLE species (taxid INT PRIMARY KEY, parent INT, rank VARCHAR(50), name VARCHAR(50))")

      print( 'INSERTING...', sep=' ', flush=True )
      for taxid in nodes_dict:
          parent = nodes_dict[ taxid ][ 'parent_tax_id' ]
          rank   = nodes_dict[ taxid ][ 'rank' ]
          name   = names_dict[ taxid ][ 'scientific name' ]
          parent,rank,name = [x.replace('"','') for x in [parent,rank,name]]
          cur.execute('INSERT INTO species VALUES('+ taxid +', '+ parent +', "'+ rank +'", "'+ name +'")')

      print( 'DONE!\n' )


  print('File tax.db created!')
      
  if not args.skip:
    ncbi_config=load_ncbi_config()
    taxdb_path=os.path.abspath(str(args.output) + 'tax.db')
    ncbi_config['ncbi_taxdb']= taxdb_path
    printerr(f"Setting ncbi_taxdb = {taxdb_path} into the user configuration --> ~/.ncbi_config", 1)
    save_ncbi_config()
    
    # if os.path.isfile(userfile) and not args.overwrite:
    #     old_content=open(userfile).readline().strip()
    #     if old_content!=content:
    #         a=None
    #         while not a in ['', 'Y', 'N']:
    #             a=input(f'\nExisting: {userfile} with content: {old_content}\nOverwriting with new content: {content}\nConfirm? (Y or Enter to proceed, N to cancel) > ')
    #         if a=='N':
    #             sys.exit(1)
    # with open(userfile, 'w') as fw:
    #     print(f'Writing user config file --> {userfile}')
    #     fw.write(content+'\n')

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from string import *
import sys
from subprocess import *
from .MMlib3 import *
from Bio import Entrez
from time import sleep
from .search_ncbi_entries import main as run_search_ncbi_entries
from .ncbi_lib import *


help_msg="""Search or fetch entries (species or lineages) from the ncbi taxonomy database through internet and print information for each hit. It wraps functions from Bio->Entrez.

ncbi_taxonomy  [define taxids || text search]   [other options]

## Define taxids (alternative, mutually exclusive):
-i   +     define inputfile  with taxids, one per line
-I  +,+,.. define taxonomy taxids, comma separated. 

## or search database:
-S  "+"    search a single string into the database, print information for each hit

### Options:
# these options allow to print additional information. See examples at the bottom
-r         print rank 
-c         print genetic code id
-m         print mitochondrial genetic code
-n         print folder-like name (with chars masked as {chXX} and spaces replaced by underscores)

-L         legacy mode. Some taxids were merged into others by ncbi. When requesting these, normally the entry for the new taxid is reported and a warning printed to stderr. If this option is active, the old id is used in output instead, so that there would be a perfect correspondance between what was asked for and what is in output (expect for taxids not found at all)
-D         deep search: if no result if found with the string provided (-S), it tries to trim it to get a partial result. E.g.  Homo sapiens mdeuslls  -> it will ignore mdeuslls and return the entry for Homo sapiens. 
-s         print summary instead of full lineage (condensate usual output to have it smaller)
-a     +   max attempts numbers. Sometimes unstability of network results in failed attempts to reach ncbi. The program tries by default this number of times before giving up. There's a wait time of 1 sec between attempts. Set -a to 0 to never give up. Careful: if this number is high (or 0), it means it takes a long time before it realizes there's a problem, like: no internet connection.
-b         batch size. default is 200, but it is lowered by 15% if the following error is found: IOError("Requested URL too long (try using EPost?)") 
-v         verbose mode
-email          email address to use with NCBI Entrez. Required for 1st usage of any online query
-h OR --help    print this help and exit

## Output:
ID : lineage (; separated) -> Scientific name [ additional fields (# separated) ]

### Example:   ncbi_taxonomy.py -S "human" -r -c -m -n 
9606 : cellular organisms; Eukaryota; Fungi/Metazoa group; Metazoa; Eumetazoa; Bilateria; Coelomata; Deuterostomia; Chordata; Craniata; Vertebrata; Gnathostomata; Teleostomi; Euteleostomi; Sarcopterygii; Tetrapoda; Amniota; Mammalia; Theria; Eutheria; Euarchontoglires; Primates; Haplorrhini; Simiiformes; Catarrhini; Hominoidea; Hominidae; Homininae; Homo -> Homo sapiens # Rank: species # GeneticCode: 1=Standard # MitoGeneticCode: 2=Vertebrate Mitochondrial # MaskedName: Homo_sapiens

If used as a function (imported in another python program), the main() function must be called with a opt dictionary as argument. opt['silent'] should be 1. An dictionary is returned. For each entry found, its taxid is a key in the dictionary, the corresponding value is the string that would be printed (so, possibly affected by options -r, -c, -m)
"""

command_line_synonyms={}

def_opt= { 
'i':0, 'I':'', 'L':0,
'S':"", 'c':0, 'm':0, 'r':0, 's':0, 'D':0, 'n':0,
'v':0, 'Q':0,
'b':200, 'silent':0, 'a':100,
'email':''
}



def lineage_string_to_abstract(lineage):
  """ lineage is a string which is usually returned by this program. This function condensate it keeping the most interesting classes. """
  splt=lineage.split('; ')
  if "Bacteria; " in lineage  :    return 'B; '+'; '.join(splt[2:min(5, len(splt))])
  elif 'Archaea; ' in lineage :    return 'A; '+'; '.join(splt[2:min(5, len(splt))])
  elif 'Eukaryota; ' in lineage:   
    out='E; '
    if 'Metazoa; ' in lineage:
      out+='M; '
      if 'Deuterostomia; ' in lineage:
        out+='Deuterostomia; '
        if 'Vertebrata; ' in lineage:
          out+='Vertebrata; '      
          if 'Mammalia; ' in lineage:           out+='Mammalia; '
          elif 'Sauropsida; ' in lineage:       out+='Sauropsida; '
          elif 'Amphibia; ' in lineage:         out+='Amphibia; '
          elif 'Actinopterygii; ' in lineage:   out+='Actinopterygii; '
          elif 'Chondrichthyes; ' in lineage:   out+='Chondrichthyes; '
        elif 'Tunicata; ' in lineage:          
          out+='Tunicata; '
          if 'Ascidiacea; ' in lineage:           out+='Ascidiacea; '         
        elif 'Branchiostomidae; ' in lineage:  out+='Branchiostomidae; '
        elif 'Echinodermata; ' in lineage:     out+='Echinodermata; '
      elif 'Protostomia; ' in lineage:
        out+='Protostomia; '
        if  'Arthropoda; ' in lineage:      
          out+='Arthropoda; '
          if 'Insecta; ' in lineage:        out+='Insecta; '
          elif 'Crustacea; ' in lineage:    out+='Crustacea; '
          elif 'Myrapoda; ' in lineage:     out+='Myrapoda; '
          elif 'Arachnida; ' in lineage:     out+='Arachnida; '
          elif 'Merostomata; ' in lineage:     out+='Merostomata; '
        elif 'Nematoda; ' in lineage:     out+='Nematoda; '
        elif 'Mollusca; ' in lineage:           
          out+='Mollusca; '
          if 'Gastropoda; ' in lineage: out+='Gastropoda; '
          elif 'Bivalvia; ' in lineage: out+='Bivalvia; '
        elif 'Annelida; ' in lineage:           out+='Annelida; '
        else:      out+=     lineage.split('Protostomia; ')[1].split(';')[0]+'; '
      else: #basal metazoan
        if 'Cnidaria; ' in lineage:        out+='Cnidaria; '
        elif 'Porifera; ' in lineage:      out+='Porifera; '          
        elif 'Ctenophora; ' in lineage:    out+='Ctenophora; '
        elif 'Placozoa; ' in lineage:      out+='Placozoa; '
        elif 'Platyhelminthes; ' in lineage: out+='Platyhelminthes; '

    else:      out+= '; '.join(splt[2:min(4, len(splt))])+'; '
    return out[:-2]
  else:      return '; '.join(splt[0:min(4, len(splt))])
            
  
#########################################################
###### start main program function

def verbose(msg):
  if opt['v']:      print(msg)

def id_is_included_in_list(main_id, id_list):
  for a_id in id_list:
    if main_id in a_id:
      return True
  return False


def main(args={}):
#########################################################
############ loading options
  global opt
  if not args: opt=command_line(def_opt, help_msg, 'io', synonyms=command_line_synonyms )
  else:  
    opt=args
    for k in def_opt: 
      if k not in opt: opt[k]=def_opt[k]
    
  #determining input ids
  id_list=[]
  if opt['i']:
    check_file_presence(opt['i'], 'input_file_with_ids')
    for line in open(opt['i']):       id_list.append(line[:-1])
  elif opt['I']:    id_list=str(opt['I']).split(',')
  elif not opt['S']:  raise Exception("ERROR must provide ids either with option -i FILE or option -I id1,id2,...,idN, or a search string with opt -S.  Run with option -h for help")

  email_setup(opt['email'])
  
  ########## search string
  if opt['S']:
    if "_"in opt['S'] and not ' ' in opt['S']:      opt['S']=replace_chars(opt['S'], '_', ' ')
    opt['S']=unmask_characters(opt['S'])
    search_ncbi_entries_opt={'silent':1, 's':opt['S'], 'm':'T', 'a':opt['a']}
    id_list= run_search_ncbi_entries(search_ncbi_entries_opt)    
      
    # searching again changing characters such as "_", ":"
    if not id_list:
      search_ncbi_entries_opt['s']=replace_chars(opt['S'], '_', ' ')
      id_list= run_search_ncbi_entries(search_ncbi_entries_opt)
    if not id_list:
      search_ncbi_entries_opt['s']=replace_chars(opt['S'], ':', ' ')
      id_list= run_search_ncbi_entries(search_ncbi_entries_opt)
    if not id_list:
      search_ncbi_entries_opt['s']=replace_chars(opt['S'], '()', ' ')
      id_list= run_search_ncbi_entries(search_ncbi_entries_opt)
    
    if opt['D'] and not id_list:
      # approximate match      
      search_string= opt['S'];       search_string= ' '.join(search_string.split()[:len(search_string.split())-1])
      while not id_list and search_string:
        search_ncbi_entries_opt={'silent':1, 's':search_string, 'm':'T', 'a':opt['a']}
        id_list= run_search_ncbi_entries(search_ncbi_entries_opt)
        if not id_list: search_string= ' '.join(search_string.split()[:len(search_string.split())-1])        

      if not search_string:
        # approximate match  replacing _ with " "
        search_string= replace_chars(opt['S'], '_', ' ');       search_string= ' '.join(search_string.split()[:len(search_string.split())-1])
        while not id_list and search_string:
          search_ncbi_entries_opt={'silent':1, 's':search_string, 'm':'T', 'a':opt['a']}
          id_list= run_search_ncbi_entries(search_ncbi_entries_opt)
          if not id_list: search_string= ' '.join(search_string.split()[:len(search_string.split())-1])        

      if id_list and search_string: printerr('WARNING returning the approximate match reducing: "'+opt['S']+'"  to:  "'+search_string+'"', 1)
      
  verbose('tot hits: '+str(len(id_list)))
  #batch preparation
  batch_size=opt['b']
  batch_index=0
  id_retrieved_until_now=0

  output_hash={}

  while id_retrieved_until_now!=len(id_list):
    try:
      batch_id_list=id_list[id_retrieved_until_now:id_retrieved_until_now+batch_size]
      if len(batch_id_list)!=len(id_list) and not opt['silent']:
        verbose('Batch n'+str(batch_index+1)+' --- entries:'+str(len(batch_id_list)))
      
      number_of_records=0
      batch_id_list_index=0

      #### fetching!
      fetch_successful=False
      error_occurred=None
      attempts=0
      while not fetch_successful and (  attempts < opt['a'] or opt['a']==0):
        try:
          fetch_handle=Entrez.efetch(db='taxonomy', id=','.join(batch_id_list), retmode='xml')    ## connecting to ncbi via internet
          parsed_results=[r for r in Entrez.parse(fetch_handle)]
          fetch_successful=True
        except KeyboardInterrupt:          raise
        except Exception as error:
          service('Something failed (network problem?). Trying again... attempt number: '+str(attempts)+ '(will keep indefinitely'*int(opt['a']==0)+int(opt['a']!=0)*' (will stop at attempt n.'+str(opt['a'])+')')
          sleep(1)
          attempts+=1
          error_occurred=error
          
      if not fetch_successful:
        printerr("ERROR couldn't fetch results. Number of attempts: "+str(attempts)+'. (Last) exception was: ', 1)
        raise error_occurred

      ###### parsing fetch results
      for record in parsed_results:
        ### note: batch_id_list[batch_id_list_index]   is the taxid we asked for
        out_taxid= record['TaxId']   #string

        while batch_id_list_index < len(batch_id_list) and batch_id_list[batch_id_list_index] != out_taxid :  #### entries should be in the same order in which we asked for. thus we check sequentially if taxid correspond. we enter here if it does not
          if 'AkaTaxIds' in record  and  batch_id_list[batch_id_list_index] in record['AkaTaxIds']:
            #### the id requested was merged into something else, this is why it doesn't match
            if opt['L']: 
              if not opt['silent']:    printerr("WARNING! Reporting entry " +batch_id_list[batch_id_list_index]+" instead of newer: " + out_taxid+' since option -L is active (species: '+record['ScientificName']+')', 1)
              out_taxid = batch_id_list[batch_id_list_index]  ### legacy mode, using the taxid you requested even if it's not the most recent one at ncbi
            else:  
              if not opt['silent']:    printerr("WARNING! Entry " +batch_id_list[batch_id_list_index]+" was merged into this: " +  out_taxid+' | '+record['ScientificName'], 1)
              batch_id_list[batch_id_list_index]=out_taxid
          else:
            ##### id was not found!!
            if not opt['silent']:  printerr("WARNING! Entry " +batch_id_list[batch_id_list_index]+" was not found and was ignored", 1)
            #printerr(    "size batch_id_list ", len(batch_id_list),  'batch_id_list_index', batch_id_list_index,   1)
            output_hash [batch_id_list[batch_id_list_index]]=''
            batch_id_list_index+=1

        
        if batch_id_list[batch_id_list_index]==out_taxid:
          lineage=record['Lineage']
          if opt['s']:   lineage=lineage_string_to_abstract(lineage)
          out=out_taxid+' : '+lineage+' -> '+record['ScientificName']

          #adding additional fields
          if opt['r']:           out+= ' # Rank: '+str(record['Rank'])
          if opt['c']:           out+= ' # GeneticCode: '+str(record['GeneticCode']['GCId'])+'='+record['GeneticCode']['GCName']
          if opt['m']:           out+= ' # MitoGeneticCode: '+str(record['MitoGeneticCode']['MGCId'])+'='+record['MitoGeneticCode']['MGCName']
          if opt['n']:           out+= ' # MaskedName: '+ mask_characters(record['ScientificName']).replace( ' ', '_' )

        
          if not opt['silent']: print(out)
          output_hash [batch_id_list[batch_id_list_index]]=out

        #if batch_id_list_index == len(batch_id_list): return output_hash
        batch_id_list_index+=1
        number_of_records+=1

      if batch_id_list_index<len(batch_id_list): 
        for i in range(batch_id_list_index,len(batch_id_list)):  
          if not opt['silent']: printerr("WARNING! entry " +batch_id_list[i]+" was not found and was ignored", 1)
          output_hash [batch_id_list[batch_id_list_index]]=''
          
      id_retrieved_until_now+=len(batch_id_list)
      batch_index+=1
    except IOError as e:
      if e.args and 'Requested URL too long' in e.args[0]:
        batch_size=int(batch_size*0.85)
        if not opt['silent']: printerr('requested URL too long... trying to decrease batch size to:'+str(batch_size), 1)
      else:
        raise e
          

  return output_hash
#######################################################################################################################################

def close_program():
  if 'temp_folder' in globals() and is_directory(temp_folder):
    bbash('rm -r '+temp_folder)
  try:
    if get_MMlib_var('printed_rchar'): 
      printerr('\r'+printed_rchar*' ' ) #flushing service msg space       
  except:
    pass


if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 



from string import *
import sys
from .MMlib3 import *
from Bio import Entrez
from urllib.error import URLError
from ftplib import FTP
import gzip
import shutil
from time import strptime, sleep
import traceback
import easyterm
from pathlib import Path
Entrez.email = None

ncbi_config=easyterm.commandlineopt.CommandLineOptions({
  'ncbi_taxdb':'',
  'ncbi_email':'',
  })

def load_ncbi_config():
  #nonlocal ncbi_config
  ncbi_config_file=Path( str(Path.home())+'/.ncbi_config' )
  if ncbi_config_file.is_file():
    ncbi_config.update(
      easyterm.read_config_file(str(ncbi_config_file))
      )
  return ncbi_config

def save_ncbi_config():
  #nonlocal ncbi_config
  ncbi_config_file=Path( str(Path.home())+'/.ncbi_config' )
  ncbi_config.write_config_file(str(ncbi_config_file))


def email_setup(email):
  ncbi_config=load_ncbi_config()
  if not email and not ncbi_config['ncbi_email']:
    raise Exception('ERROR on your first usage, you must provide an email address to use NCBI services with -email')

  if email and email!=ncbi_config['ncbi_email']:
    printerr(f"Setting email = {email} into the user configuration --> ~/.ncbi_config", 1)
    ncbi_config['ncbi_email']=email
    try:
      save_ncbi_config()
    except:
      printerr('WARNING: could not save configuration to ~/.ncbi_config', 1)

  Entrez.email=email


    
#### NOTE some values must be defined in MMlib.opt:
# sleep_time
# retmax
# max_attempts
set_MMlib_var('opt', options( {'retmax':250, 'max_attempts':10, 'sleep_time':5, 'v':0}))  ## this is to make this lib work without MMlib command line calls

""" Wrapping default Entrez methods to connect to ncbi to allow network problems and batch requests"""
def esearch(**keyargs):
  """ Generic wrap. You should use term=..  and db=..  and field=..
  Note: these will be used:  opt['max_attempts']   opt['sleep_time']; opt['retmax']
   where opt is get_MMlib_var('opt')   [you can set it with  set_MMlib_var('opt', opt) ]
  Returns a list of uids (strings) """
  opt=get_MMlib_var('opt');  max_attempts=opt['max_attempts']; sleep_time=opt['sleep_time']; retmax=opt['retmax']
  uid_list=[]; retstart=0
  while not uid_list or (int(searched_g['Count']) > len(uid_list) ):  ## to manage when more results are returned than the allowed number
    if  uid_list:    retstart+=int(searched_g['RetMax'])  ## same as setting it to len(genome_uid_list)
    ### add stuff for network problems
    n_attempt=0 #-1 means success
    while n_attempt < max_attempts:
      try:                
        verbose('ESEARCH.n{3} retstart:{0} retmax:{1} {2}'.format(retstart, retmax, ' '.join([k+':'+str(keyargs[k]) for k in keyargs]), n_attempt+1), 1)
        searched_g = Entrez.read( Entrez.esearch(retstart=retstart, retmax=retmax,  **keyargs), validate=False ) 
        break
      except URLError:    sleep(sleep_time); n_attempt+=1;  service('esearch FAILED attempt n{0} trying again in {1}s'.format(n_attempt, sleep_time) )
    if n_attempt == max_attempts: write('esearch ERROR network problem??', 1); raise
    ###### 
    #write(searched_g, 1, how='green')
    if not int(searched_g['Count']):  break
    verbose('ESEARCH-OUTIDS: {0}'.format( ' '.join(searched_g['IdList'])), 1 )
    uid_list.extend( searched_g['IdList'] )
  assert int(searched_g['Count']) == len(uid_list)  ## control to see that code is written properly
  return uid_list

def efetch(**keyargs):
  """ Generic wrap. You should use db=..  and id=..list_of_ids..  
  in here, id MUST Be a list (Entrez.efetch accepts a single id as well)
  The fetch is performed by with rettype='docsum'.
  Returns a list of the objects parsed with Entrez.read; the fields of the objects depend on the db queried
  Normally we expect the same quantity of objects in input and in output, a control is performed.
 """
  opt=get_MMlib_var('opt');  max_attempts=opt['max_attempts']; sleep_time=opt['sleep_time']; retmax=opt['retmax']
  batch_size=int(retmax)
  ### add stuff
  list_out=[]
  if not 'id' in keyargs   or  type(keyargs['id'])!=list:  raise Exception("efetch ERROR id must be provided and must be a list! keyargs: {0} ".format(keyargs))
  if not keyargs['id']: return []
  for batch_index in range( 1+  (len(keyargs['id'])-1) // batch_size ):
    batch_this_list = keyargs['id'][ batch_size*batch_index : batch_size*(batch_index+1) ]
    if not batch_this_list: continue # bad design, inherited. sorry!
    #write( (batch_index, len(batch_this_list)), 1, how='blue' )
    this_keyargs= dict(keyargs);   del this_keyargs['id']
    ### add stuff for network problems
    n_attempt=0
    while n_attempt < max_attempts:
      try:              
        verbose('EFETCH.n{2} {1} idlist: {0} '.format(' '.join([b for b in batch_this_list]), ' '.join([k+':'+str(keyargs[k]) for k in keyargs if k!='id']), n_attempt+1 ), 1)
        parsed  =Entrez.read( Entrez.efetch(id=batch_this_list, rettype='docsum', validate=False, **this_keyargs), validate=False )
        break
      except URLError:    sleep(sleep_time); n_attempt+=1; service('efetch FAILED attempt n{0} trying again in {1}s'.format(n_attempt, sleep_time) )
    if n_attempt == max_attempts: write('efetch ERROR network problem??', 1); raise

    if    issubclass(type(parsed),list):    fetched_obj_list = parsed
    elif  'DocumentSummarySet' in parsed:   fetched_obj_list = parsed['DocumentSummarySet']['DocumentSummary']
    else:   raise Exception('efetch ERROR I do not know how to handle this object to get a list of results! {0}'.format(parsed))
    if len(fetched_obj_list) != len(batch_this_list): raise Exception("efetch ERROR some ids could not be found! idlist: {0} \n{1}".format(' '.join([b for b in batch_this_list]), '\n'.join([str(o) for o in fetched_obj_list])))
    list_out.extend(fetched_obj_list)
  return list_out

ftp_handler=None
def get_ftp_handler(url='ftp.ncbi.nlm.nih.gov', force=False):
  global ftp_handler
  if ftp_handler is None or force:
    verbose('FTP: establishing connection with {}'.format(url), 1)
    ftp_handler=FTP(host=url)
    ftp_handler.login()
  return(ftp_handler)

### copy-pasted from hurry.filesize:
size_names = [    (1024 ** 5, 'P'),    (1024 ** 4, 'T'),     (1024 ** 3, 'G'),     (1024 ** 2, 'M'),     (1024 ** 1, 'K'),    (1024 ** 0, 'B')    ]
def human_readable_size(bytes, system=None):   
  """Human-readable file size.
  Using the traditional system, where a factor of 1024 is used::     >>> size(10)    '10B'
  >>> size(20000)    '19K'       >>> size(100000)    '97K'           >>> size(200000)    '195K'        """
  the_size_names=size_names if system is None else system
  for i, (factor, suffix) in enumerate(the_size_names):
    if bytes >= factor:            break
  amount = int(bytes/factor)
  if system is None and amount>=1000:
    amount=1
    suffix=the_size_names[i+1][1]
  return str(amount) + suffix


#def wget_spider(ffile, return_size=False):
#  """ given a file path in a ftp server, checks with wget if it exists """
#  while True:
#    if ffile=='None': return '   '
#    wget_command='wget --spider {0} '.format(ffile); verbose(wget_command, 1); log = bash(wget_command)[1]
#    splt=log.split()
#    if   splt and splt[-1]=='exists.' :                                 
#      if not return_size:      return 'web'
#      else: 
#        for index, sstring in enumerate(splt):
#          if sstring=='SIZE': return int(splt[index+3])
#        raise Exception, 'couldnt parse size here! \n'+log
#
#    elif len(splt)>3 and join(splt[-4:-1], ' ')=='No such file' :       return '---'    
#    else:  
#      raise Exception,  'wget_spider ERROR with file: "{0}" \nThis is the wget log:{1}'.format(ffile, log)



#! /usr/bin/env python3
from .ncbi_lib import *
from .MMlib3 import options
help_msg=""" Displays information about ncbi db that can be queried through the Entrez module. Usage:

$  ncbi_db_info    db_name

To get a list of possible db_name, run with no options

### Options:
-email         email address to use with NCBI Entrez. Required for first usage of ncbi online tools
-h OR --help   print this help and exit"""

command_line_synonyms={}

def_opt= { 'email':'',
'i':0, }


#########################################################
###### start main program function

def main(args=None):
#########################################################
############ loading options
  global opt

  if args is None:
    opt=command_line(def_opt, help_msg, 'i', synonyms=command_line_synonyms )
  else:
    opt=options(def_opt)
    opt.update(args)

  email_setup(opt['email'])
  
  set_MMlib_var('opt', opt)
  #global temp_folder; temp_folder=Folder(random_folder(opt['temp'])); test_writeable_folder(temp_folder, 'temp_folder'); set_MMlib_var('temp_folder', temp_folder)
  #global split_folder;    split_folder=Folder(opt['temp']);               test_writeable_folder(split_folder); set_MMlib_var('split_folder', split_folder) 
  #checking input
  input_name= opt['i']

  if not input_name:
    ## no input specified, let's see the list of dbs
    write('These is the list of possible DBs available at ncbi:', 1)
    write(' '+   '\n '.join (  sorted(Entrez.read( Entrez.einfo() )['DbList'])), 1  )

  else:
    write( 'Name'.ljust(5)+' '+ 'FullName'.ljust(35)  +' '+  'Description', 1)  
    for k in sorted(Entrez.read( Entrez.einfo(db=input_name) )['DbInfo']['FieldList'], key=lambda x:x['FullName']):
      write( k['Name'].ljust(5)+' '+ k['FullName'].ljust(35)  +' '+  k['Description'], 1)


  ###############



#######################################################################################################################################

def close_program():
  # if 'temp_folder' in globals() and is_directory(temp_folder):
  #   bbash('rm -r '+temp_folder)
  try:
    if get_MMlib_var('printed_rchar'): 
      printerr('\r'+printed_rchar*' ' ) #flushing service msg space       
  except:
    pass

  if 'log_file' in globals(): log_file.close()


if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 



#!/usr/bin/env python3
from string import *
import sys
from subprocess import *
from .MMlib3 import *
from .ncbi_taxonomy import main as run_ncbi_taxonomy
from .ncbi_lib import email_setup
from Bio import Entrez, SeqIO
import time

help_msg="""Program to fetch entries from the ncbi sequence databases through the internet and print information such as its sequence or the source organism. It wraps Bio->Entrez and Bio->SeqIO.

ncbi_sequences  [mode] [ids] [options]

## Define mode (which ncbi database is queried):
-m  +      Possibilities: N -> nucleotide, P -> protein

## Define IDs; anything that ncbi understand is accepted (e.g. AAB88790.1):
-i  +      define inputfile  with IDs, one per line
-I  +[,+]  define IDs on the command line, comma separated

## Define output: (if multiple output is defined, they are all printed)
-M         minimal header: only the primary accession id is printed
-L         one entry per line: no newlines are printed, tab separators are printed between output information
-f         fasta sequence output
-o         species output 
-t         taxonomy+species output (this is slower than just -o, since an additional web query is performed)
-x         output all available fields, but do not output sequence

The output contains the desired fields (fasta, species, taxonomy) in this order. So if you asked for fasta and taxonomy for example ( -f -t ) you will have: 

>AAB88790 selenophosphate synthetase [Drosophila melanogaster]
7227 : cellular organisms; Eukaryota; Opisthokonta; Metazoa; Eumetazoa; Bilateria; Coelomata; Protostomia; Panarthropoda; Arthropoda; Mandibulata; Pancrustacea; Hexapoda; Insecta; Dicondylia; Pterygota; Neoptera; Endopterygota; Diptera; Brachycera; Muscomorpha; Eremoneura; Cyclorrhapha; Schizophora; Acalyptratae; Ephydroidea; Drosophilidae; Drosophilinae; Drosophilini; Drosophilina; Drosophiliti; Drosophila; Sophophora; melanogaster group; melanogaster subgroup -> Drosophila melanogaster
MSYAADVLNSAHLELHGGGDAELRRPFDPTAHDLDASFRLTRFADLKGRGCKVPQDVLSK
LVSALQQDYSAQDQEPQFLNVAIPRIGIGLDCSVIPLRHGGLCLVQTTDFFYPIV

### Options:
-email +         email address to use with NCBI Entrez. Required for first usage of ncbi online tools
-a     +         max attempts numbers. Sometimes unstability of network results in failed attempts to reach ncbi. The program tries by default this number of times before giving up. There's a 1 sec wait between attempts. Set -a to 0 to never give up. Careful: if this number is high (or 0), it means it takes a long time before it realizes there's a problem, like: no internet connection.
-w     +         compulsory waiting time between NCBI connections
-b               batch size. default is 50, but it is lowered by 15% if the following error is found: IOError("Requested URL too long (try using EPost?)") 
-print_opt       print currently active options
-h OR --help     print this help and exit

### When used as function imported in another python program, the "main" function must be run using the opt hash which you would use to get the information, and using the opt['silent']=1.  Example:
from ncbi_db import ncbi_sequences; out_hash=ncbi_sequences.main({'I':'AAB88790', 'f':1, 'silent':1})
An out_hash is returned,   with as key the id provided and as value a list [full_id, seq_if_requested, taxonomy_+_species_if_requested]. Full_id is minimal if option M is active
"""

command_line_synonyms={}

def_opt= {'temp':'/tmp/', 
          'm':'', 'a':100,
          'i':0, 'I':'',
          'M':0, 'L':0,
          'f':0, 't':0, 'o':0,
          'v':0, 'Q':0, 'x':0,
          'w':0.5,
          'b':50, 'silent':0,
          'email':'',
}

mode_synonyms={'N':'nucleotide', 'P':'Protein'}

#########################################################
###### start main program function

def main(args={}):
#########################################################
  def id_is_included_in_list(main_id, id_list):
    for a_id in id_list:
      if main_id in a_id:
        return True
    return False


############ loading options
  global opt
  if not args: opt=command_line(def_opt, help_msg, 'io', synonyms=command_line_synonyms )
  else:  
    opt=args
    for k in def_opt: 
      if k not in opt: opt[k]=def_opt[k]
      
  def verbose(msg):
    if opt['v']:       printerr( msg, 1)
  #global temp_folder=Folder(random_folder(opt['temp'])); test_writeable_folder(temp_folder, 'temp_folder'); set_MMlib_var('temp_folder', temp_folder)
  #global split_folder;    split_folder=Folder(opt['temp']);               test_writeable_folder(split_folder); set_MMlib_var('split_folder', split_folder) 


  #mode
  if not opt['m']:  raise Exception("ERROR you must specify a mode with option -m (N|P). See --help")
  try:
    db=mode_synonyms[opt['m'].upper()] 
  except: raise Exception("ERROR mode not recognized: "+opt['m'])
  opt['m']=opt['m'].upper()
  #determining input ids
  id_list=[]
  if opt['i']:
    check_file_presence(opt['i'], 'input_file_with_ids')
    for line in open(opt['i']):       id_list.append(line[:-1])
  elif opt['I']:    id_list=str(opt['I']).split(',')
  else: raise Exception("ERROR must provide ids either with option -i FILE or option -I id1,id2,...,idN. Run with option -h for help")
  #output preparing
  line_sep='\n'
  if opt['L']: line_sep='\t'
  #batch preparation
  batch_size=opt['b']
  batch_index=0
  id_retrieved_until_now=0
  headers_checklist={} ### to deal with duplicate fetching

  output_hash={}

  email_setup(opt['email'])
  
  while id_retrieved_until_now!=len(id_list):
    try:
      batch_id_list=id_list[id_retrieved_until_now:id_retrieved_until_now+batch_size]
      if len(batch_id_list)!=len(id_list) and not opt['silent']:
        verbose('Batch n'+str(batch_index+1)+' --- entries:'+str(len(batch_id_list)))
      
      #retrieving! Seq_accver': 'CM000066.5', u'TSeq_sequence
      rettype='gb'
      if opt['m']=='N' and not opt['x'] : rettype='fasta'

      number_of_records=0
      batch_id_list_index=0

      fetch_successful=False
      error_occurred=None
      attempts=0
      while not fetch_successful and (  attempts < opt['a'] or opt['a']==0):
        if opt['w'] and attempts>0:  time.sleep(opt['w'])
        try:
          #print( f"db: {db} id: {','.join(batch_id_list)} retmode=xml  rettype={rettype}")
          fetch_handle=Entrez.efetch(db=db, id=','.join(batch_id_list), retmode='xml', rettype=rettype)
          parsed_results=[r for r in Entrez.parse(fetch_handle)]
          fetch_successful=True
        except KeyboardInterrupt:          raise
        except Exception as error: ##debug
          #print('Something failed (network problem?). Trying again... attempt number: '+str(attempts)+ ' (will keep trying'*int(opt['a']==0)+int(opt['a']!=0)*(' (will stop at attempt n.'+str(opt['a']))+')')            
          if not opt['silent']:
            service('Something failed (network problem?). Trying again... attempt number: '+str(attempts)+ ' (will keep trying'*int(opt['a']==0)+int(opt['a']!=0)*(' (will stop at attempt n.'+str(opt['a']))+')')
            time.sleep(1)
          error_occurred=error
          attempts+=1

      if not fetch_successful:
        printerr("ERROR couldn't fetch results. Number of attempts: "+str(attempts)+'. (Last) exception was: ', 1)
        raise error_occurred
      
      for record in parsed_results:
        if opt['m']=='N':
          #[u'TSeq_accver', u'TSeq_sequence', 'GBSeq_other-seqids', u'TSeq_length', u'TSeq_taxid', u'TSeq_orgname', u'TSeq_sid', u'TSeq_gi', u'TSeq_seqtype', u'TSeq_defline']
          #interactive_mode
          if opt['x']:
            write('#'*100, 1, how='blue')
            for k in sorted(record.keys()):
              write( str(k.ljust(25)), 0, how='green')
              write(' '+str(record[k]), 1)
              header=''
          else: 
            seq=record['TSeq_sequence'].upper()
            organism=record['TSeq_orgname']
            if 'TSeq_accver' in record:
              header=record['TSeq_accver']
            else: 
              if 'TSeq_sid' not in record: raise Exception("ERROR record format not recognized for this entry: "+str( '\n'. join([str(k)+': '+str(record[k]) for k in record])))
              header=record['TSeq_sid']

            if not opt['M']:
              ## COMPUTING NICE STRING FOR IDS
              #header=header+" gi|"+record['TSeq_gi']+'|'+header+' '+record['TSeq_defline']
              header=header+' '+record['TSeq_defline']
            if header.split()[0] in headers_checklist: continue
            headers_checklist[header.split()[0]]=True

          #print header
          
          record['GBSeq_other-seqids']=[[batch_id_list[batch_id_list_index]]]   ## to keep homogenous procedures below...

        elif opt['m']=='P':        

          if opt['x']:
            write('#'*100, 1, how='blue')
            for k in sorted(record.keys()):
              write( str(k.ljust(20)) + ' '+str(record[k]), 1)
              header=''
          if True: #
            seq=record['GBSeq_sequence'].upper()          
            organism=record['GBSeq_organism']
            #taxonomy=record['GBSeq_taxonomy']+' -> '+organism   ### ---> this is approximate and does not contain all the sublevels. also, it is not available for nucleotides, which makes the whole thing undesirable.

            #for k in record:
            #  print k+' '+str(record[k])[:100]
            #raw_input('..')
            header=  record['GBSeq_accession-version'] if 'GBSeq_accession-version' in record else    record['GBSeq_primary-accession']
            if not opt['M']:       
            ## COMPUTING NICE STRING FOR IDS
              gi_id=''; ref_id=''; other_ids=[]
              for a_id in record['GBSeq_other-seqids']:
                if    a_id.split('|')[0]=='gi': gi_id=a_id
                elif  a_id.split('|')[0]=='ref': 
                  ref_id=a_id
                else: other_ids.append(a_id)
              id_string=''
              if gi_id:   id_string+=gi_id
              if ref_id:  
                if id_string and id_string[-1]!='|': id_string+='|' #adding a | if is not already there
              id_string+=  ref_id
              for a_id in other_ids:
               if id_string and id_string[-1]!='|': id_string+='|' 
               id_string+=  a_id 
              header+=' '+id_string+' '+record['GBSeq_definition']        

          if header.split():
            if header.split()[0] in headers_checklist: continue        
            headers_checklist[header.split()[0]]=True
              
          ##########
            

         
        while batch_id_list_index < len(batch_id_list) and not id_is_included_in_list( batch_id_list[batch_id_list_index], record['GBSeq_other-seqids'] ):
          if not opt['silent']:  printerr("WARNING! entry " +batch_id_list[batch_id_list_index]+"  was not found and was ignored", 1)
          output_hash[batch_id_list[batch_id_list_index]]=[]
          batch_id_list_index+=1

        if batch_id_list_index == len(batch_id_list): sys.exit()
        output_hash[batch_id_list[batch_id_list_index]]= [header]
          
        out= '>'+header+line_sep

        if opt['o']:
          output_hash[batch_id_list[batch_id_list_index]].append(organism)        
          out+=organism+line_sep
          
        if opt['t']:
          ncbi_tax_opt={'silent':True, 'S':organism, 'a':opt['a']}
          results=run_ncbi_taxonomy( ncbi_tax_opt )
          if len(results)==0:
            if not opt['silent']: printerr("WARNING! organism " +organism+" returned no results in taxonomy! keeping this empty", 1)
            taxonomy = 'unknown'
          else:
            if len(results)>1: 
              if not opt['silent']: printerr("WARNING! organism " +organism+" returned more than a result in taxonomy! taking a random one", 1)
#            raise Exception, "ERROR searching species in ncbi_taxonomy: "+organism+" ; number of results: "+str(len(results))
            taxonomy = results[ list(results.keys())[0]  ]
  
          output_hash[batch_id_list[batch_id_list_index]].append(taxonomy)
          out+=taxonomy+line_sep
        
        if opt['f']: #fasta
          output_hash[batch_id_list[batch_id_list_index]].append(seq)
          if opt['L']:         out+=seq+line_sep
          else:                out+= fasta(seq)+line_sep

        if not opt['silent'] and not opt['x']: write(out[:-1], 1)

        batch_id_list_index+=1
        number_of_records+=1
      id_retrieved_until_now+=len(batch_id_list)

      if batch_id_list_index<len(batch_id_list): 
        for i in range(batch_id_list_index,len(batch_id_list)):  
          output_hash[batch_id_list[batch_id_list_index]]=[]
          if not opt['silent']: printerr("WARNING! entry " +batch_id_list[i]+"  was not found and was ignored", 1)

      batch_index+=1
      

    except IOError as e:
      if e.args and 'Requested URL too long' in e.args[0]:
        batch_size=int(batch_size*0.85)
        if not opt['silent']: verbose('requested URL too long... trying to decrease batch size to:'+str(batch_size))
      else: raise

  return output_hash
#######################################################################################################################################

def close_program():
  if 'temp_folder' in globals() and is_directory(temp_folder):
    bbash('rm -r '+temp_folder)
  try:
    if get_MMlib_var('printed_rchar'): 
      printerr('\r'+printed_rchar*' ' ) #flushing service msg space       
  except:
    pass


if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 



#! /usr/bin/env python3
from string import *
import sys
from subprocess import *
from .MMlib3 import *
#import gbparsy # import GBParsy module


help_msg="""Extract specific feature from a GenBank flat file, using the GBParsy module.

Usage: $ parse_genbank.py   gbff_file    [other_options]   > output_file.fa

## Options:
-f   feature to extract (default: "gene")
-q   qualifiers to add to the fasta header  (special values accepted: "all", "fasta_title", "header")
-s   filter; word to search in values of any qualifier, only if found the entry will be printed
-t   try and use this qualifier as title; comma-separated list can be provided (default: auto; which means gene->locus_tag; cds->protein_id,locus_tag)
-p   for coding sequences extraction, output peptide sequence instead of nucleotide (implies -f CDS and -q product,locus_tag)
-l   long output; qualifiers are added in the format "qualifier=value;" instead of just "value"

-x   output 'source' entries instead of features (i.e. contig sequences); disregard all other options
-d   do not output sequence; the headers will also lack the leading ">"
-g   suppress normal output; output gff format instead  

-print_opt      print currently active options
-h OR --help    print this help and exit

#### Examples:
Extract proteome:  # parse_genbank.py genbank.gbff -p      > proteome.fa
Extract genome:    # parse_genbank.py genbank.gbff -x      > genome.fa
Extract mRNA:      # parse_genbank.py genbank.gbff -f mRNA > mRNAs.fa"""

command_line_synonyms={}

def_opt= { 
'i':0, 'q':0, 's':0, 'p':0, 'd':0, 'l':0,
'f':'gene', 't':'auto', 
'x':0, 'g':0,
}


#########################################################
###### start main program function

def main(args={}):
#########################################################
############ loading options
  import gbparsy # import GBParsy module
  

  global opt
  if not args: opt=command_line(def_opt, help_msg, 'io', synonyms=command_line_synonyms )
  else:  opt=args
  set_MMlib_var('opt', opt)
  #checking input

  
  check_file_presence(opt['i'], 'Genbank gbff input file')
  sFileName = opt['i'];  sFeature = opt['f'];  
  sQualifiers = []  if not opt['q'] else opt['q'].split(',')
  sWord =      None if not opt['s'] else opt['s']
  add_fields = []   if not opt['a'] else opt['a'].split(',')
  title_fields= [] if not opt['t'] else opt['t'].split(',')
  get_protein= opt['p']
  
  if get_protein: 
    sFeature = "CDS"
    if not sQualifiers: sQualifiers=['product', 'locus_tag']

  if sQualifiers==['all']: sQualifiers='ALL'

  if title_fields ==['auto']:
    if    sFeature=='CDS':  title_fields=['protein_id','locus_tag']
    else  :                 title_fields=['locus_tag']

  

  id_dictionary={} # just to keep track of ids already used
  ids_changed_since_duplicated=0

  oGBParsy = gbparsy.gbparsy()
  service('parsing {f}'.format(f=sFileName))
  oGBParsy.parse(sFileName) # parse a GBF file which contains more than one GBF sequence data

#  printerr('something', 1)

  # If you want to get a list of SeqRecord instances, use getBioData method instead of getRawData.
  lSeqData = oGBParsy.getRawData()

  failed_parse_titles=0   
  lack_translation_warnings=0
  for dSeqData in lSeqData: # dSeqData has a parsed data of a GBF sequence data
      # start of user process
      if opt['x']:
        # for k in dSeqData:
        #   if k!='sequence':
        #     write( k, how='red')
        #     write(dSeqData[k], 1)
        # raw_input('..')
        
        accession = dSeqData['version']  if 'version' in dSeqData and dSeqData['version'] else    dSeqData['accession'] 
        source_title='{id} {desc}'.format(id=accession, desc=dSeqData['definition'])   ### you also have organism, lineage ...
        write(">{tit}\n{seq}".format(tit=source_title, seq=fasta(   dSeqData['sequence'].upper())  ),1  )
        continue

      for dFeature in dSeqData["features"]:
          if dFeature["feature"] == sFeature or sFeature=='*':
              #write(dFeature, 1)

              #for k in dSeqData.keys():
                #write( k+'   --> '+str(dSeqData[k]), 1, how='red')
                #raw_input('...')

              print_this=True if not sWord else bool( searchQualValue(sWord, dFeature) )  ## we print only if we find this word as value of any qualifiers
              title='{feat}_{start}-{end}_{strand}'.format(feat=dFeature["feature"], start=dFeature["start"], end=dFeature["end"], strand=dFeature["direction"])

              #printerr(title, 1)
              chromosome_name= dSeqData["version"] if 'version' in dSeqData else dSeqData["accession"]
              g=gene(chromosome=chromosome_name   , strand={'C':'-', 'N':'+'}[dFeature["direction"]])
              for start, end in dFeature["locations"]:          g.add_exon(start, end)
              
              for title_field in title_fields:
                q=getQualValue(title_field, dFeature)
                if not q:
                  if title_field == title_fields[-1]:             failed_parse_titles+=1
                else:                 title=q; break

              added_suffix=0
              while title in id_dictionary:
                if not added_suffix:  title=title+'-'+str(added_suffix+1)
                else:                 title=   '-'.join(title.split('-')[:-1]) +'-'+str(added_suffix+1)
                added_suffix+=1

              id_dictionary[title]=True
              if added_suffix: ids_changed_since_duplicated+=1

              g.id=title

              description=' '
              if sQualifiers != 'ALL': 
                for sQualifier in sQualifiers:                    
                  if   sQualifier == 'header':      value=g.header(no_id=True)
                  elif sQualifier == 'fasta_title': value=g.fasta_title()
                  else:                             value= getQualValue(sQualifier, dFeature)
                  if not opt['l']: description+=value+' '
                  else:            description+=sQualifier+'='+value+'; '
              else: 
                for sQualifier, value in dFeature["qualifiers"]:
                  if not opt['l']: description+=value+' '
                  else:            description+=sQualifier+'='+value+'; '

              if   opt['d']:
                write('{title}{desc}'.format(title=title, desc=description.rstrip()), 1)
              elif opt['g']:
                write(g.gff(program='Genbank', tag=dFeature["feature"], comment=description.rstrip()), 1)
              else:
                if get_protein: 
                  sequence=getQualValue('translation',  dFeature)
                  if not sequence: 
                    lack_translation_warnings+=1
                    if lack_translation_warnings<10:
                      printerr('WARNING lacking translation field for entry: {t} ; obtaining from raw sequence'.format(t=title), 1)
                    if lack_translation_warnings==10:
                      printerr('LAST WARNING: lacking translation field for entry: {t}; obtaining from raw sequence. Next warnings of this type will not be printed'.format(t=title), 1)                      
                    codon_start=getQualValue('codon_start',  dFeature)
                    cds=getSequence(dSeqData["sequence"], dFeature)
                    if codon_start and codon_start!='1':  cds=cds[int(codon_start)-1:]
                    sequence=transl( cds ).rstrip('*').rstrip('X')
                else:  sequence=getSequence(dSeqData["sequence"], dFeature) 
                write('>{title}{desc}\n{seq}'.format(title=title, desc=description.rstrip(), seq=sequence), 1)


  if failed_parse_titles: 
    printerr('WARNING: could not get desired field title for {n} entries; used a generic title instead'.format(n=failed_parse_titles), 1)

  if ids_changed_since_duplicated: 
    printerr('WARNING: {n} titles were not unique, and were changed'.format(n=ids_changed_since_duplicated), 1)

  ###############



#######################################################################################################################################

def close_program():
  if 'temp_folder' in globals() and is_directory(temp_folder):
    bbash('rm -r '+temp_folder)
  try:
    if get_MMlib_var('printed_rchar'): 
      printerr('\r'+printed_rchar*' ' ) #flushing service msg space       
  except:
    pass

  if 'log_file' in globals(): log_file.close()



sNorBase = "ACGTRYMKWSBDHVNacgtrymkwsbdhvn";
sComBase = "TGCAYRKMWSVHDBNtgcayrkmwsvhdbn";

def getRevCom(sSequence):
    lSequence = [sComBase[sNorBase.find(x)] for x in sSequence]
    lSequence.reverse()
    return "".join(lSequence)

def getQualValue(sQualifier, dFeature):
    if not sQualifier: return ""
    for tQualifier in dFeature["qualifiers"]:
        if tQualifier[0] == sQualifier: return tQualifier[1]
    return ""

def searchQualValue(sWord, dFeature):
    if not sWord: return ""
    for tQualifier in dFeature["qualifiers"]:
        if sWord in tQualifier[1]: return tQualifier[1]
    return ""

def getSequence(sWholeSequence, dFeature):
    sSequence = sWholeSequence[dFeature["start"] - 1: dFeature["end"]]
    if dFeature["direction"] == "C":        sSequence = getRevCom(sSequence)
    return sSequence


if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 



#__all__=['__version__']
import functools
from ._version import __version__

from .ncbi_assembly import main as run_ncbi_assembly
from .ncbi_sequences import main as run_ncbi_sequences
#from .ncbi_db_info import main as run_ncbi_db_info
from .ncbi_pubmed import main as run_ncbi_pubmed
from .ncbi_search import main as run_ncbi_search
from .ncbi_taxonomy import main as run_ncbi_taxonomy
from .ncbi_taxonomy_tree import main as run_ncbi_taxonomy_tree
from .parse_genbank import main as run_parse_genbank

def suppress_return(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        f(*args, **kwargs)
        return None
    return wrapper

run_ncbi_assembly=suppress_return(run_ncbi_assembly)
run_ncbi_sequences=suppress_return(run_ncbi_sequences)
run_ncbi_pubmed=suppress_return(run_ncbi_pubmed)
run_ncbi_search=suppress_return(run_ncbi_search)
run_ncbi_taxonomy=suppress_return(run_ncbi_taxonomy)
run_ncbi_taxonomy_tree=suppress_return(run_ncbi_taxonomy_tree)
run_parse_genbank=suppress_return(run_parse_genbank)




#!/usr/bin/env python3
import sys,os
import json
import sqlite3
import argparse
from collections import defaultdict
from .MMlib3 import bbash, unmask_characters, mask_characters, get_taxids_from_ncbi_db, printerr, random_folder
import shutil, os
from easyterm import NoTracebackError
from .ncbi_taxonomy_tree_db import main as run_ncbi_taxonomy_tree_db
from .ncbi_lib import load_ncbi_config, save_ncbi_config
from pathlib import Path

import warnings
warnings.filterwarnings("ignore", category=SyntaxWarning)
from ete3 import Tree

def get_parser():
    """returns the argparse parser object with the argument option requirements"""

    parser     = argparse.ArgumentParser(description='Get the phylogenetic tree from a list of species (based on NCBI taxonomy)')
    taxid_file = parser.add_argument_group('input')
    tree       = parser.add_argument_group('output')
    render     = parser.add_argument_group('render')
    database   = parser.add_argument_group('database')

    parser.add_argument('-temp', help='temporal folder', default='/tmp/')
    parser.add_argument('-q','--quiet', help='do not show warnings', action="store_true", default=False)

#    taxid_file.add_mutually_exclusive_group(required = True)
    taxid_file.add_argument("-n", "--name",
                        help="list of scientific names, one name per line. Only scientific names from NCBI Taxonomy (not synonyms!)\
                              are recognized")
    taxid_file.add_argument("-t", "--taxid",
                        help="list of taxonomy ids, one taxid per line. The scientific name in NCBI associated to the taxid is used\
                              in the tree")
    taxid_file.add_argument("-u", "--unmask",
                        help="allow masked characters in the input file", action="store_true")
    taxid_file.add_argument("-m", "--mask",
                        help="include masked characters in the tree", action="store_true")
    taxid_file.add_argument("-tn","--taxid_name",
                        help="two columns input file: taxid<TAB>species_name. species_name is used in the tree")
    taxid_file.add_argument("-ta","--taxid_annotation",
                        help="include any attribute(s) as a feature in any of the nodes of the tree. The input file format is\
                              'taxid=value[;key=value]' ( taxid key is required). The node corresponding to the specified taxid\
                               (including internal nodes) will include each key=value as a feature")
    taxid_file.add_argument("-na","--name_annotation",
                        help="same as --taxid_annotation, but including name instead of taxid: 'name=value[;key=value]'")
    tree.add_argument("-f","--fix_internal",
                      help="internal nodes present in the input list will be added a sister leaf node with the same name and taxid",
                      action='store_true')
    tree.add_argument("-N","--normalize",
                      help="normalize the number of ranks.", action='store_true')
    tree.add_argument("-l","--ladderize",
                      help="sort the branches of a given tree (swapping children nodes) according to the size of each partition.", action='store_true')
    tree.add_argument("-um","--ultrametric",
                      help="converts a tree to ultrametric topology (all leaves must have the same distance to root).", action='store_true')
    tree.add_argument("-nw","--newick",
                      help="output tree in newick format")
    tree.add_argument("-js","--json",
                      help="output tree in json format")
    tree.add_argument("-x",
                      help="output file with taxids list")
    tree.add_argument("--lineage",
                      help="get the full lineage for each species. Each terminal node includes a 'lineage' attribute", action="store_true")
    tree.add_argument("-nn","--no_name",
                      help="do not include species name in the leaves, use taxid instead", action="store_true")
    tree.add_argument("-d", "--non_dicotomic",
                      help="keep non_dicotomic nodes", action="store_false")
    tree.add_argument("-s", "--show_tree",
                      help="show interactive tree", action="store_true")
    tree.add_argument("-p", "--print_stdout",
                      help="print tree to stdout", action="store_true")
    
    render.add_argument("-r","--render",
                        help="render tree as image. File extension will determine the image format (.PDF, .SVG or .PNG)")
    render.add_argument("-H","--height",
                        help="height of the image in units",
                        type=float,
                        default=None)
    render.add_argument("-W","--width",
                        help="width of the image in units",
                        type=float,
                        default=None)

    render.add_argument("-U","--units",
                        help="px: pixels (default), mm: millimeters, in: inches",
                        default='px')
    render.add_argument("-DPI",
                        help="dots per inches (300)",
                        type=float,
                        default=300)
    database.add_argument('-db', '--database',
                          help="tax.db file. See --makedb to obtain it. If not specified, reads path from ~/.ncbi_taxdb if existing",
                          default=None)
    database.add_argument('-mdb', '--makedb',
                        help="Create a tax.db file. For a dedicated help page, run: ncbi_taxonomy_tree --makedb -h",
                          default=None)

    return parser


def parse_opts(x):
    """to be used when the program is imported as a module:
import ncbi_taxonomy_tree
args = ncbi_taxonomy_tree.parse_opts('-y your -o options -f for -t the -p program')
ncbi_taxonomy_tree.main(args)
"""
    parser = get_parser()                         
    args = parser.parse_args(x.split())
    return args


def get_taxid_from_names(input_names, temp_dir, unmask, database, is_list=False, quiet=False):
    """ gets a file with the list of names, and returns the list of taxids and a dict {taxid:name}"""
    if not is_list:
        names_list = set()
        for line in open(input_names):
            if unmask:
                names_list.add( unmask_characters(line.strip().replace('_',' ')) )
            else:
                names_list.add(line.strip())
    else:
        if not input_names: raise Exception('ERROR if is_list=True, arg is mandatory')
        names_list = set(input_names)
    print_stderr( '%s unique names in input\n' % (len(names_list)), quiet)
    names_db = '/'.join( database.split('/')[:-1] + ['names.dmp'] )
    if not os.path.exists(names_db):
        raise Exception('ERROR file not found: '+ names_db)
    out_hash, ambigous_hash, not_found_hash = get_taxids_from_ncbi_db(names_list, ncbi_db=names_db, temp_dir=temp_dir, full=True)
    if ambigous_hash: print_stderr(  'WARNING: %s ambiguous names: %s\n' % (len(ambigous_hash),' '.join( ambigous_hash )) , quiet)
    if not_found_hash: print_stderr( 'WARNING: %s names not found in ncbi: %s\n' % (len(not_found_hash), ','.join(list(not_found_hash.keys()))) , quiet)
    # check if there is a correspondence 1:1 between names and taxids
    taxids_from_input_dict={}
    for k,v in list(out_hash.items()):
        taxids_from_input_dict.setdefault(v,[]).append(k)
    for x,v in list(taxids_from_input_dict.items()):
        if len(v)>1:
            print_stderr( 'WARNING: %s elements in your input returned the same taxid, only the first one will be included in the tree: %s\n' % (len(v),', '.join(v)) , quiet)
    return out_hash,taxids_from_input_dict

def get_taxid_and_annotation():
    """ """
    taxid_list = set()
    if not os.path.exists(args.taxid_annotation): raise Exception('file not found: %s\n' % args.taxid_annotation)
    infile = args.taxid_annotation
    taxid_attrs = {}
    for line in open(infile):
        if not 'taxid=' in line:
            raise Exception('ERROR: taxid key is missing in '+ line.strip() +'. Please, use "taxid=value[;key=value]" in the input file.') 
        attrs_dict = {}
        for attr in line.strip().split(';'):
            k,v = attr.split('=')
            if 'taxid' in k:
                taxid = int(v)
                continue
            attrs_dict[ k ] = v
        taxid_attrs[ taxid ] = attrs_dict
        taxid_list.add(taxid)
    return list(taxid_list), taxid_attrs

def get_name_and_annotation(arg, temp_dir):
    """ """
    if not os.path.exists(arg): raise Exception('file not found: %s\n' % arg)
    name_attrs = {}
    input_names = []
    for line in open(arg):
        if not 'name=' in line:
            raise Exception('ERROR: taxid key is missing in "'+ line.strip() +'". Please, use "name=value[;key=value]" in the input file.')
        attrs_dict = {}
        for attr in line.strip().split(';'):
            k,v = attr.split('=')
            if 'name' in k:
                if args.unmask:
                    name = unmask_characters(v.replace('_',' '))
                else:
                    name = v
                continue
            attrs_dict[ k ] = v
        name_attrs[ name ] = attrs_dict
        input_names.append(name)
    out_hash, taxids_from_input_dict = get_taxid_from_names(input_names, temp_dir, args.unmask, args.database, is_list=True, quiet=args.quiet)
    taxid_list = list(out_hash.values())
    # map taxid and attributes
    taxid_attrs = {}
    for name,taxid in list(out_hash.items()):
            taxid_attrs[ taxid ] = name_attrs[ name ]
    return taxid_list, taxids_from_input_dict, taxid_attrs

def node_lineage(node, t):
    """DEPRECATED! returns the lineage of node as a list with the ancestral nodes. each element in the list corresponds to (distance, rank, name), where distance is the number of nodes between
 node and the root of the tree."""
    dist = int(node.get_distance(t, topology_only=True))
    lineage = [ (dist,node.rank,node.name,node.taxid) ]
    parent = node.up
    while parent and not parent.name == 'Root':
        dist = int(parent.get_distance(t, topology_only=True))
        lineage.append( (dist,parent.rank,parent.name,parent.taxid) )
        parent = parent.up
    return sorted( lineage, key = lambda x:x[0] )

def include_lineage(t):
    """ """
    lineages_dict={}
    for node in t.get_descendants('preorder'):
        if not node.is_leaf():
            for name in node.get_leaf_names():
#                lineages_dict.setdefault(name,[]).append( (node.rank,node.name,node.taxid) )
                lineages_dict.setdefault(name,[]).append( node.name )
        else:
            node.add_feature('lineage', '||'.join( lineages_dict[node.name] + [node.name] )   )


def print_stderr(msg, quiet=False):
    if not quiet:
        sys.stderr.write(msg)


def main(args=None):

    if args is None:

        if '--makedb' in sys.argv or '-mdb' in sys.argv:
            sys.argv = [x for x in sys.argv if not x in ['--makedb', '-mdb']]
            run_ncbi_taxonomy_tree_db()
            sys.exit()

        parser = get_parser()
        args = parser.parse_args()

    ranks = ['superkingdom', 'kingdom', 'phylum', 'class', 'order', 'family', 'genus', 'species']


    # ncbi database file
    ncbi_config=load_ncbi_config()    
    if args.database is None:
        # userfile=str(Path.home())+'/.ncbi_taxdb'
        # if os.path.exists(userfile):
        #     with open(userfile) as fr:
        #         content=fr.readline().strip()
        #args.database=content.split('ncbi_taxdb')[1].split('=')[1].strip()
        args.database=ncbi_config['ncbi_taxdb']
        if not args.database:
            args.database=None #just to ensure type is consistent
    elif not args.database is None and not ncbi_config['ncbi_taxdb']:
        taxdb_path=os.path.abspath(args.database)
        ncbi_config['ncbi_taxdb']=taxdb_path
        printerr(f"Setting ncbi_taxdb = {taxdb_path} into the user configuration --> ~/.ncbi_config", 1)        
        save_ncbi_config()
        
    if args.database is None or not os.path.exists(args.database):
        raise NoTracebackError('ERROR: --database file not found: %s\nPerhaps you never built a NCBI tax.db? For help, run: ncbi_taxonomy_tree --makedb -h' % args.database)

    temp_folder=random_folder(args.temp).rstrip('/')+'/'
    
    #### INPUT
    # if input file contains taxonomy ids
    if args.taxid:
        if not os.path.exists(args.taxid): raise Exception('file not found: %s\n' % args.taxid)
        taxid_list = list(set([x.rstrip() for x in open(args.taxid)]))
        try:
            [int(x) for x in taxid_list ]
        except:
            raise Exception('ERROR: this does not look like a taxonomy id: %s; try --name instead\n' % x)

    # if input file contains names
    elif args.name:
        if not os.path.exists(args.name): raise Exception('file not found: %s\n' % args.name)
        out_hash,taxids_from_input_dict = get_taxid_from_names(args.name, temp_folder, args.unmask, args.database, quiet=args.quiet)
        taxid_list = list(out_hash.values())

    # if both taxid and name in input
    elif args.taxid_name:
        if not os.path.exists(args.taxid_name): raise Exception('file not found: %s\n' % args.taxid_name)
        infile = args.taxid_name
        taxid_name_dict = {}
        taxid_list = set()
        for line in open(infile):
            taxid, name = line.strip().split("\t")
            taxid_name_dict[ taxid ] = name
            taxid_list.add(taxid)

    # if an annotation is provided for the taxids
    elif args.taxid_annotation:
        taxid_list, taxid_attrs = get_taxid_and_annotation()

    # if an annotation is provided for the names
    elif args.name_annotation:
        taxid_list, taxids_from_input_dict, taxid_attrs = get_name_and_annotation(args.name_annotation, temp_folder)

    # list of taxids is ready for building the tree
    if not taxid_list:
        sys.stderr.write('no taxids to build the tree, probably your input does not match with ncbi names,\
 please try --unmask option, in case your input names are masked\n')
        sys.exit()

    set_taxid=set(taxid_list)
    taxid_list = list(set_taxid)


    #### BUILD
    print_stderr( 'input for tree: ' + str(len(taxid_list)) + ' taxids\n' , args.quiet)
    print_stderr( 'building tree...', args.quiet)
    nodes_dict = {}
    taxids_remove = []
    cursor = sqlite3.connect(args.database).cursor()
    t, nodes_dict, taxids_remove = buildTree(taxid_list, nodes_dict, taxids_remove, cursor)
    t.add_feature('rank', 'Root')
    print_stderr( 'DONE!\n\n', args.quiet)
    if taxids_remove:
        print_stderr('WARNING: the following ids were not found in the taxonomy database used: '+ ','.join(map(str,taxids_remove)) +'\n', args.quiet)

    for n in t.traverse():
        n._requested= n.taxid in set_taxid
        
    ### add control: if requested, do not trim 
        
    #### OUTPUT
    # get the full lineage of species, this is done before removing non-dicotomic nodes, so all nodes are presenent (no need for -d option)
    if args.lineage:
        include_lineage(t)


    # taxonomy ids file
    if args.x:
        open(args.x, 'w').write('\n'.join( map(str, taxid_list) )+'\n')
        print_stderr( str(len(taxid_list)) +' taxids stored in file: '+ args.x +'\n', args.quiet)            

    # normalize ranks
    if args.normalize:
        t=normalize_ranks(t, ranks)

    # remove non_dicotomic nodes
    if args.non_dicotomic and not args.normalize:
        if len(t.get_children()) == 1:
            t.get_children()[0].delete(prevent_nondicotomic=False)
        for node in t.get_descendants():
#            if len(node.get_children()) == 1 and not node.taxid in taxid_list:
            if len(node.get_children()) == 1 and not node.taxid in set_taxid:            
#                node.delete(prevent_nondicotomic=True)
                node.delete(prevent_nondicotomic=False)                

                
    # include node annotation
    if args.taxid_annotation or args.name_annotation:
        for taxid, attrs_dict in list(taxid_attrs.items()):
            try:
                node = t.search_nodes(taxid=taxid)[0]
                node.add_features(**attrs_dict)
            except:
                # problems detected so far:
                # two taxids in the list point to the same node, then only one is present in the tree and the other one is missing (although it was not include in the taxids_remove list because it returned a result)
                print_stderr("WARNING: "+str(taxid)+' not found in the tree, its attributes could not be included\n', args.quiet)

    # json outfile (node.name is still the taxid)
    if args.json:
        if args.ladderize:
            t.ladderize()
        jsoned = ete2json(t)
        with open(args.json, 'w') as json_outfile:
            json.dump(jsoned, json_outfile)
        print_stderr('Tree saved in json format: '+ args.json +'\n', args.quiet)


    # include species name
    internal_nodes_in_input=[]
    if not args.no_name:
        if args.name:
            for taxid,names_list in list(taxids_from_input_dict.items()):
                nodes = t.search_nodes(taxid = taxid)
                if len(nodes)>1: raise Exception('ERROR: more than one node with taxid: %s' % taxid)
                if not nodes:  raise Exception('ERROR: no node has taxid: %s' % taxid)
                nodes[0].add_features(name=names_list[0])
                if not nodes[0].is_leaf(): internal_nodes_in_input.append( nodes[0] )
        elif args.taxid_name:
            for leaf in t.get_leaves():
                leaf.name = taxid_name_dict[ leaf.taxid ]
        else:
            for leaf in t.get_leaves():
                leaf.name = leaf.name


    if args.fix_internal:# and internal_nodes_in_input:
        for taxid in taxid_list:
            
            nodes = t.search_nodes(taxid = taxid)
            if len(nodes)>1: raise Exception('ERROR: more than one node with taxid: %s' % taxid)
            if not nodes:
                sys.stderr.write( 'WARNING: no node has taxid: %s\n' % taxid )
                continue
            if nodes[0].is_leaf(): continue
            node = nodes[0]
            print_stderr('MESSAGE: internal node duplicated as a leaf: %s\n' % node.name , args.quiet)
            new_node = node.up.add_child(name='artificial_node')
            new_node.add_feature('rank','no rank')
            d = node.detach()
            new_node.add_child(d)
            new_leaf = new_node.add_child(name=node.name)
            new_leaf.add_features( taxid=node.taxid, rank='no rank' ) #scientific_name=node.scientific_name,
            if len(node.get_children()) == 1: node.delete(prevent_nondicotomic=False)



            
    print_stderr( '\n'+ str(len(taxid_list)) +' taxids in input\n'+ str(len(t)) +' leaves in tree\n', args.quiet)


    if args.mask:
        for leaf in t.get_leaves():
            leaf.name = mask_characters( leaf.name ).replace(' ','_')
        
    if args.ladderize:
        t.ladderize()

    if args.ultrametric:
        t.convert_to_ultrametric(t.get_farthest_node()[1])

    # newick outfile
    if args.newick:
        t.write(outfile=args.newick, features=[])
        print_stderr('Tree saved in newick format: '+ args.newick +'\n')
    
    if args.render:
        t.render(args.render, w=args.width, h=args.height, units=args.units, dpi = args.DPI)
    
    if args.show_tree:
        t.show()

    if args.print_stdout:
        print(t)

    if os.path.isdir(temp_folder):
        shutil.rmtree(temp_folder)
    return t

# END OF PROGRAM
# FUNCTIONS
def normalize_ranks(t,ranks):
    normtree = Tree()
    for l in t.get_leaves():
        out = []
        lineage={}
        taxids={}
        parent = l.up
        while parent:
            if parent.rank in ranks:
                lineage[parent.rank]=parent.scientific_name
                taxids[parent.rank]=parent.taxid
            parent = parent.up

        for rank in ranks:
            if not rank in lineage:
                lineage[rank] = 'No '+rank
                taxids[rank]= 'No taxid'

        for i in range(len(ranks)):
            rank=ranks[i]
            if i==0: # superkingdom
                if not normtree.search_nodes(name=lineage[rank], rank=rank):
                    child = normtree.add_child(name=lineage[rank])
#                    child.add_features(rank=rank,lineage=lineage[rank],taxid=taxids[rank])
                    child.add_feature('rank', rank )
                    child.add_feature('lineage', lineage[rank])
                    child.add_feature('taxid', taxids[rank])
            else:
                if rank == 'species' and lineage[rank] == 'No species':
                    lineage['species'] = l.name
                parent_node_lineage =':'.join([ lineage[x] for x in ranks[:i]])
                node_lineage =':'.join([ lineage[x] for x in ranks[:i] + [rank]])
                parent = normtree.search_nodes(lineage=parent_node_lineage)
                if not parent or len(parent)>1:
                    raise
                if not normtree.search_nodes(name=lineage[rank], lineage=node_lineage):
                    child = parent[0].add_child(name=lineage[rank])
#                    child.add_features(rank=rank,lineage=lineage[rank],taxid=taxids[rank])
                    child.add_feature('rank', rank )
                    child.add_feature('lineage', node_lineage)
                    child.add_feature('taxid', taxids[rank])
                    if rank == 'species':
                        child.scientific_name = l.scientific_name
                        child.taxid = l.taxid
    return normtree


#def node_lineage(node):
#    """returns lineage of node"""
#    lineage = [ node.rank ]
#    parent = node.up
#    while parent and not parent.name == '1':
#        lineage.insert(0,parent.rank)
#        parent = parent.up
#    return lineage

def collapse_species(t, taxid_list, ranks):
    """collapse internal nodes if they are present in input list.
this is meant to solve the situation in which a taxid has children taxids also present in the list.
Once solved, the resulting tree has each taxid from the list as a leaf.
It is inconsistennt with their lineages, but this step is necessary for its representation in the sunburst plot.
"""
    stop = False
    for node in t.get_descendants():
        if not node.is_leaf() and node.name in taxid_list:
            
            if node.rank in ranks:
                print_stderr('ERROR: taxid "'+ node.name +'" is present in infile and has children also present in infile.\nIt can not be collapsed because its rank "'+ node.rank +'" is among the items in ranks list '+ str(ranks) +'. Please, remove it from the infile.\n')
                print_stderr( str(len(list(set(node.get_leaf_names()).intersection(taxid_list)))) +' leaves names under this node: ' + str(list(set(node.get_leaf_names()).intersection(taxid_list))) +'\n', args.quiet)
                stop = True
                
            for child in node.get_descendants('postorder'):
                if child.name in taxid_list:
                    detached = child.detach()
                    node.up.add_child(detached)
                else:
                    child.detach()

    if stop: # this tag allows to identify all the problematic nodes in one run
        sys.exit()

#def normalize_ranks(t, input_ranks):
#    """WORKING ON IT...
#gets a tree and normalizes the number of ranks.
#the returned tree will contain only the ranks in the ranks list,
#the same number of ranks for all the species.
#"""
#    sys.stderr.write('MESSAGE: normalizing ranks in tree to ' + ','.join(input_ranks) +'\n')
#    ranks = ['root'] + input_ranks[:-1]
#    rank_node_down_dict = {} # {rank:rank_down} 
#    for r in range(len(ranks)-1):
#        rank_node_down_dict[ ranks[r] ] = ranks[r+1]
#
#    t.rank = 'root'
#
#    for node in t.get_descendants('postorder'):
#        if node.is_leaf():
#            ## Normalize leaves
#            ## I want the parent of the leaf to be the last element in the ranks list
#            ## so I need to collapse everything under that rank
#            
#            lineage = node_lineage(node)
#
#            ## check if duplicate ranks in lineage
#            lineage_copy = lineage[:]
#            while 'no rank' in lineage_copy:
#                lineage_copy.remove('no rank')
#            if not len(lineage_copy) == len(set(lineage_copy)):
#                sys.stderr.write('ERROR: lineage with duplicated ranks. Please, remove taxid: '+ node.name + ' from input file.\nInconsistent lineage: '+ str(lineage) +'\n')
#                sys.exit()
#
#            if not node.up.rank == ranks[-1]:
#                ## parent rank is not last item in ranks list
#                parent = node.up
#                if ranks[-1] in lineage:
#                    ## last item in ranks list is present in lineage
#                    ## delete nodes until rank of parent is ranks[-1]
#                    while not parent.rank == ranks[-1]:
#                        parent.delete(prevent_nondicotomic=False)
#                        parent = node.up
#
#                else:
#                    ## ranks[-1] is not present in lineage
#                    ## I will reconstruct the lineage
#
#                    ## nodes with rank not present in ranks will be removed
#                    ## until the rank of is present in ranks
#                    while not parent.rank in ranks:
#                        child = parent
#                        parent = parent.up
#                        child.delete(prevent_nondicotomic=False)
#                        
#                    while not parent.rank == ranks[-1]:
#                        new = parent.add_child(name='und')
#                        new.rank = rank_node_down_dict[ parent.rank ]
#                        new.scientific_name = ''
#                        child = node.detach()
#                        new.add_child(child)
#                        parent = new
#        else:
#            ## node is not leaf
#            ## I will simply delete the node if its rank is not in ranks
#            if not node.rank in ranks:
#                node.delete(prevent_nondicotomic=False)
#                
#    ## include missing ranks
#    for node in t.get_descendants('levelorder'):
#        if not node.is_leaf():
#            while not node.rank == rank_node_down_dict[ node.up.rank ]:
#                new = node.up.add_child(name='und')
#                new.rank = rank_node_down_dict[ node.up.rank ]
#                new.scientific_name = ''
#                child = node.detach()
#                new.add_child(child)
#                
#    for node in t.get_leaves():
#        node.rank = 'species'
#        if not node_lineage(node) == input_ranks:
#            sys.stderr.write('WARNING: There is a problem with taxid: '+ node.name +'\n' )
#            sys.stderr.write('         lineage != input_ranks...\n')
#            sys.stderr.write('         lineage: '+ str(node_lineage(node)) +'\n')
#            sys.stderr.write('         input_ranks: '+ str(input_ranks) +'\n')


def annotate_tree(t,infile):
    """Annotates tree from infile (key=value[;key=value]). taxid key required. node.name must be the taxid."""
    taxid_attrs = defaultdict()
    for line in open(infile):
        if not 'taxid=' in line:
            sys.stderr.write('taxid key is missing in '+ line.strip() +'. Please, use "taxid=value;" in the input file.\n')
            sys.exit()
        else:
            attrs_dict = {}
            for attr in line.strip().split(';'):
                k,v = attr.split('=')
                attrs_dict[ k ] = v
            taxid_attrs[ attrs_dict['taxid'] ] = attrs_dict
            
    for taxid, attrs_dict in list(taxid_attrs.items()):
        try:        node = t&taxid
        except:
            sys.stderr.write("ERROR "+str(taxid)+' not found! ')
            raise 
        node.add_features(**attrs_dict)

    
    
def ete2json(t, jsoned = dict()):
    """Recursive function to convert Tree into serializable json object"""
    if not t.is_leaf():
        jsoned[ 'children' ] = []
        for node in t.get_children():
            children = {}
            children['taxid'] = node.name
            children['rank']  = node.rank
            children['name']  = node.scientific_name
            children = ete2json(node, children)
            jsoned['children'].append(children)
    else:
        jsoned['size'] = '1'
    return jsoned

def query_a_list(taxid_list, cursor):
    """gets a list of taxids, splits it in batches of size 999,
and does the query to the ncbi database local file.
Returns a list where each item has (taxid, parent_taxid, rank, name)"""
    n = 999
    results = []
    for i in range(0, len(taxid_list), n):
        batch = taxid_list[i:i+n]
        placeholders = ', '.join(['?']*len(batch))
        cur_query = 'SELECT * FROM species WHERE taxid IN (%s)' % placeholders
        cursor.execute(cur_query, batch)
        results += cursor.fetchall()
    return results


def buildTree(taxid_list, nodes_dict, taxids_remove, cursor):
    """Recursive function, returns a ete tree object from a list of taxids.
Requires a cursor connected to a sqlite db build using the script /users/rg/didac/NCBI/Taxonomy/update_sqlite_DB.py
nodes_dict is an empty dict
taxids_remove is an empty list """
    results = query_a_list(taxid_list, cursor)
    
    # check if all taxids returned a result
    if len(set(taxid_list)) != len(results):
        taxids_with_result = set([ x[0] for x in results])
        taxids_remove += list(set(map(int, taxid_list)) - taxids_with_result )

    parent_taxid_list = []
    for result in results:
        taxid, parent_taxid, rank, name = result
        parent_taxid_list.append(parent_taxid)

        if not taxid in nodes_dict:
            c = Tree()
            c.add_feature('name', name)
            nodes_dict[ taxid ] = c

        # I don't have scientific name and rank for parent_taxid yet, but next iteration it will be the taxid
        nodes_dict[ taxid ].add_features(name=name, taxid=taxid, rank=rank)
        # add child to node parent_taxid
        if not parent_taxid in nodes_dict:
            p = Tree()
            p.add_feature('taxid', parent_taxid)
            p.add_child( nodes_dict[ taxid ] )
            nodes_dict[ parent_taxid ] = p

        else:
            # check if taxid is a child of parent_taxid (already in nodes_dict), otherwise adding it
            for descendant in nodes_dict[ parent_taxid ].iter_descendants():
                if taxid == descendant.taxid:
                    break
            else:
                nodes_dict[ parent_taxid ].add_child( nodes_dict[ taxid ] )

    parent_taxid_list = list(set(parent_taxid_list))

    try:
        # "1" is the root of the NCBI tree, if "1" is in parent_taxid_list, and it will become an empty list inside this try
        parent_taxid_list.remove(1)
    except:
        pass

    if parent_taxid_list:
        t,nodes_dict,taxids_remove = buildTree(parent_taxid_list, nodes_dict, taxids_remove, cursor)
    else:
        nodes_dict[ 1 ].add_features(name='Root', rank='Root')
        return nodes_dict[ 1 ], nodes_dict, taxids_remove
    return t, nodes_dict, taxids_remove

      
if __name__ == '__main__':
    main()




#! /usr/bin/env python3
from .ncbi_lib import *
help_msg="""Utility to search pubmed entries online and get formatted reference-style entries 
Usage:
$ ncbi_pubmed [search] [output]

###  Search options:
-P   pubmed id(s), comma-separated  |  -p   pubmed ids file
-kp  citationKey -tab- pubmed id file; output will be in style citationKey -tab- citation

##  instead of providing pubmed ids you can combine any of these:
-S   default search; this is like standard pubmed search, you can also include field names e.g. "Nature [JOUR]"
-T   title search
-A   author; use double quotes to include name or initial; comma-separate multiple authors
-D   publication date (year only is accepted)

###  Output options:
#    Reference style:
-s  [style]   reference style is in the specified format. Available: pnas, elife, info
-ma max authors listed (if different than default for style)
#    Format options: 
-f  [format]  Use html and open with a browser to have formatted text. Available: html, txt
#    Misc:
-sa  sort per citation_in_text
-ct  text only output: id|citationInText    
     where citationInText is like "Mariotti et al., 2015" or "Chapple and Guigo, 2008" and id is the pubmed id or the citationKey if -kp

### other options:
-email        email address to use with NCBI Entrez. Required for first usage of ncbi online tools
-print_opt    print currently active options
-h OR --help  print this help and exit"""

command_line_synonyms={}

def_opt= { 
'P':0, 'p':0, 'kp':0,
's':'info',
'f':'txt',
'ma':0,
'S':0, 'T':0, 'A':0, 'D':0,
'retmax':250, 'max_attempts':10, 'sleep_time':5, 
'o':0,
'ct':0, 'sa':0,
'v':0,
'email':'',
}

#########################################################
###### start main program function

def main(args={}):
#########################################################
############ loading options
  global opt
  if not args: opt=command_line(def_opt, help_msg, 'io', synonyms=command_line_synonyms )
  else:  opt=args
  set_MMlib_var('opt', opt)
  #global temp_folder; temp_folder=Folder(random_folder(opt['temp'])); test_writeable_folder(temp_folder, 'temp_folder'); set_MMlib_var('temp_folder', temp_folder)
  #global split_folder;    split_folder=Folder(opt['temp']);               test_writeable_folder(split_folder); set_MMlib_var('split_folder', split_folder) 
  #checking input

  available_output_formats={'html':1, 'txt':1}
  output_format=opt['f']
  if not output_format in available_output_formats: raise Exception('ERROR invalid argument to option -f !  See -h')

  available_ref_style={'pnas':1, 'info':1, 'elife':1}
  ref_style=opt['s']
  if not ref_style in available_ref_style: raise Exception('ERROR invalid argument to option -s !  See -h')

  email_setup(opt['email'])

  style2max_authors={'pnas':5, 'elife':30, 'info':0} #not used for info
  max_authors= opt['ma'] if opt['ma'] else style2max_authors[ref_style]
  
  ###   determining pubmed ids
  if opt['p'] or opt['P']:
    pubmed_ids= str(opt['P']).split(',') if opt['P'] else  [line.strip() for line in open(opt['p']) if line.strip() ]
  elif opt['kp']:
    pubmed_id2citation_key={}
    for line in open(opt['kp']):
      if not line.strip(): continue
      splt=line.strip().split('\t')
      citation_key, pubmed_id= splt
      pubmed_id2citation_key[pubmed_id]=citation_key
    pubmed_ids=list(pubmed_id2citation_key.keys())
    #print pubmed_id2citation_key
  else: 
    term=''
    if opt['T']: 
      for i in opt['T'].split(): term+= ''+i +'[TITL] '
    if opt['A']: 
      for i in opt['A'].split(','): term+= ''+i.strip() +'[AUTH] '
    if opt['D']: 
      term+= ' '+str(opt['D']) +'[PDAT] '
    if opt['S']: term+= ''+ opt['S'] +' ' #[ALL] '
    #write(term, 1, how='red')
    if not term:
      raise Exception("ERROR no article search defined. See -h")
    
    if opt['v']:  printerr( 'search term: ' +term, 1)
    pubmed_ids=esearch(db='pubmed', term=term)

  ### fetching entries from ncbi
  entries=   efetch(db='pubmed', id=pubmed_ids)
  if not entries: 
    printerr('No entries found!', 1)
    sys.exit(9)

  ## building output
  textout=''
  if output_format=='html':  textout+='<html><head><title>ncbi_pubmed output</title></head><body>'

  def get_misc(entry): 
      all_authors=entry['AuthorList']
      year=entry['PubDate'].split()[0]  #so_bit.split()[0];     
      author_summary=''
      citation_in_text=''
      for auth_index, author in enumerate(all_authors):
        splt=author.split()
        surname=''; name_bit=''
        for splt_index, splt_bit in enumerate(splt):
          if splt_bit.upper() == splt_bit: name_bit+=splt_bit+', '
          else:                           surname+=splt_bit+' '
        
        surname=surname.strip()
        name_bit=name_bit.strip();         name_bit=name_bit.strip(','); 
        author_summary+=   surname + ' '+name_bit  + ', '

        if    auth_index==0: citation_in_text+=surname
        elif  auth_index==1:
          if len(all_authors)==2: citation_in_text+=' and '+surname
          else:                   
            #if citation_in_text[-2:]==', ':             citation_in_text=citation_in_text[:-2]
            citation_in_text+= ' et al.'

        if len(all_authors)>max_authors: 
          if author_summary[-2:]==', ':             author_summary=author_summary[:-2]
          author_summary+=' et al. '; break
      author_summary=author_summary.strip().strip(',')
      citation_in_text+=', '+str(year)
      return author_summary, citation_in_text

  if opt['sa']:    entries.sort(key=lambda x:get_misc(x)[1])

  for entry in entries:
    #print entry
    pubmed_id=entry['Id'];   doi=entry['DOI'] if 'DOI' in entry else 'None'
    journal_brief=entry['Source']; journal_full=entry['FullJournalName']
    pages=entry['Pages']
    ref_count=entry['PmcRefCount'];   
    all_authors=entry['AuthorList']
    so_bit=entry['SO'];     
    year=entry['PubDate'].split()[0]  #so_bit.split()[0];     
    where_bit=';'.join(so_bit.split(';')[1:])
    title=entry['Title']
    pubstatus=entry['PubStatus'] 
    ## common ?
    if pubstatus=='aheadofprint' and not where_bit:   where_bit='Published online {date} doi:{doi}'.format(date=entry['EPubDate'], doi=doi)
    author_summary, citation_in_text =  get_misc(entry) 

    #### opt -kp
    citk=pubmed_id2citation_key[pubmed_id]+' ' if opt['kp'] else ''
      
    if opt['ct']: 
      if opt['kp']: textout+=pubmed_id2citation_key[pubmed_id]+'|'+citation_in_text+'\n'
      else:         textout+=pubmed_id+'|'+citation_in_text+'\n'
      continue ### no other output!


    if ref_style == 'info':
      textout+='\n### PUBMED ID: '+pubmed_id+'\n'
      for k in sorted(entry.keys()):
        try: val=str(entry[k])
        except: val=entry[k].encode('ascii', errors='replace')
        textout+=  k.ljust(20)+' : '+val  +'\n'

    else:

      if ref_style =='pnas':

        if output_format=='txt':
          textout+='{citk}{aut} ({year}) {title} {jour} {where}\n'.format(aut=author_summary, year=year, title=title, jour=journal_brief, where=where_bit, citk=citk)
        elif output_format=='html':
          textout+='<p>{citk}{aut} ({year}) {title} <i>{jour}</i> {where}</p>'.format(aut=author_summary, year=year, title=title, jour=journal_brief, where=where_bit, citk=citk)

      elif ref_style =='elife':
        splt=author_summary.split(', ')
        first_author=splt[0];     rest_authors=', '.join(splt[1:])
        if rest_authors: rest_authors=', '+rest_authors
        splt=where_bit.split(':')
        first_where=splt[0];     rest_where=':'.join(splt[1:])
        if '(' in first_where: first_where=first_where.split('(')[0]
        if rest_where: rest_where=':'+rest_where
        doi_bit=' doi: '+doi+'.' if doi !='None' else ''

        if output_format=='txt': pass
          #textout+='{citk}{aut} ({year}) {title} {jour} {where}\n'.format(aut=author_summary, year=year, title=title, jour=journal_brief, where=where_bit, citk=citk)

        elif output_format=='html':
          ### formatted
          #textout+=u'<p>{citk}<b>{faut}</b>{raut}. {year}. {title} <i>{jour}</i> <b>{fwhere}</b>{rwhere}. <font color="blue">{doi}</font>.</p>'.format(faut=first_author, raut=rest_authors, year=year, title=title, jour=journal_brief, fwhere=first_where, rwhere=rest_where, doi=doi_bit,  citk=citk)
          textout+='<p>{citk}{faut}{raut}. {year}. {title} <i>{jour}</i> <b>{fwhere}</b>{rwhere}.{doi}</p>'.format(faut=first_author, raut=rest_authors, year=year, title=title, jour=journal_brief, fwhere=first_where, rwhere=rest_where, doi=doi_bit,  citk=citk)
        

    

  if output_format=='html':  textout+='</body>'

  if output_format=='txt':
    print(textout.encode('utf8', errors='replace').decode('utf8')  )
  elif output_format=='html':
    print(textout.encode('ascii', 'xmlcharrefreplace').decode('ascii'))

  ###############



  



#######################################################################################################################################

def close_program():
  if 'temp_folder' in globals() and is_directory(temp_folder):
    bbash('rm -r '+temp_folder)
  try:
    if get_MMlib_var('printed_rchar'): 
      printerr('\r'+printed_rchar*' ' ) #flushing service msg space       
  except:
    pass

  if 'log_file' in globals(): log_file.close()


if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 
